/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: boardlist_view.c,v 1.35.2.6 2005/02/18 19:12:44 fuyu Exp $
 */

#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"

#include "ochusha_ui.h"
#include "boardlist_view.h"
#include "marshal.h"

#include "ugly_gtk2utils.h"

#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <gtk/gtk.h>

#if DEBUG_WIDGET || DEBUG_WIDGET_MOST
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>


static void boardlist_view_class_init(BoardlistViewClass *klass);
#if GTK_MINOR_VERSION <= 2
static void popup_menu_callback(gpointer data, guint action,
				GtkWidget *widget);
static void popup_menu_callback2(gpointer data, guint action,
				 GtkWidget *widget);
#else
static void popup_menu_open_cb(GtkAction *action, GtkUIManager *manager);
static void popup_menu_open_board_in_tab_cb(GtkAction *action,
					    GtkUIManager *manager);
static void popup_menu_open_board_with_web_browser_cb(GtkAction *action,
						      GtkUIManager *manager);
static void popup_menu_toggle_hidden_states_cb(GtkAction *action,
					       GtkUIManager *manager);
static void popup_menu_kill_items_cb(GtkAction *action, GtkUIManager *manager);
static void popup_menu_copy_board_url_cb(GtkAction *action,
					 GtkUIManager *manager);
static void popup_menu_open_properties_cb(GtkAction *action,
					  GtkUIManager *manager);
#endif

static void popup_menu_destructed_cb(gpointer data);
static void boardlist_view_init(BoardlistView *view);

static void boardlist_view_finalize(GObject *object);
static void boardlist_view_destroy(GtkObject *object);

static void boardlist_view_do_on_selection(BoardlistView *view,
					   int what_to_do);
static void boardlist_view_do_on_item_at_cursor(BoardlistView *view,
						BoardlistViewAction what_to_do);

static void tree_view_row_expanded_cb(GtkTreeView *tree_view,
				      GtkTreeIter *iter, GtkTreePath *path,
				      BoardlistView *view);
static void tree_view_row_collapsed_cb(GtkTreeView *tree_view,
				       GtkTreeIter *iter, GtkTreePath *path,
				       BoardlistView *view);
static gboolean boardlist_view_enter_notify(GtkWidget *widget,
					    GdkEventCrossing *event);
static gboolean boardlist_view_button_press(GtkWidget *widget,
					    GdkEventButton *event);
static gboolean boardlist_view_button_release(GtkWidget *widget,
					      GdkEventButton *event);
static void decorate_boardlist_entry(GtkTreeViewColumn *tree_column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *tree_model,
				     GtkTreeIter *iter, gpointer unused);
static void append_board(gpointer data, gpointer user_data);
static void append_category(gpointer data, gpointer user_data);


GType
boardlist_view_get_type(void)
{
  static GType blv_type = 0;

  if (blv_type == 0)
    {
      static const GTypeInfo blv_info =
	{
	  sizeof(BoardlistViewClass),
	  NULL, /* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)boardlist_view_class_init,
	  NULL, /* class_finalize */
	  NULL, /* class_data */
	  sizeof(BoardlistView),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)boardlist_view_init,
	};

      blv_type = g_type_register_static(GTK_TYPE_TREE_VIEW,
					"BoardlistView", &blv_info, 0);
    }

  return blv_type;
}


GType
boardlist_view_action_get_type(void)
{
  static GType action_type = 0;
  if (action_type == 0)
    {
      static const GEnumValue values[] =
	{
	  { TOGGLE_HIDDEN_STATE,
	    (char *)"TOGGLE_HIDDEN_STATE", (char *)"toggle-hidden-state" },
	  { KILL_ITEM,
	    (char *)"KILL_ITEM", (char *)"kill-item" },
	  { ADVANCE_VIEW,
	    (char *)"ADVANCE_VIEW", (char *)"advance-view" },
	  { BACK_VIEW,
	    (char *)"BACK_VIEW", (char *)"go-back-view" },
	  { 0, NULL, NULL },
	};
      action_type = g_enum_register_static("BoardlistViewAction", values);
    }
  return action_type;
}


typedef struct _CategoryNodeInfo
{
  GtkTreeRowReference *reference;
} CategoryNodeInfo;


enum {
  OPEN_THREADLIST_VIEW_SIGNAL,

  SELECT_CATEGORY_SIGNAL,
  CATEGORY_COLLAPSED_SIGNAL,
  CATEGORY_EXPANDED_SIGNAL,

  TOGGLE_HIDE_CATEGORY_SIGNAL,
  TOGGLE_HIDE_BOARD_SIGNAL,
  KILL_CATEGORY_SIGNAL,
  KILL_BOARD_SIGNAL,
  COPY_BOARD_URL_SIGNAL,
  OPEN_PROPERTIES_SIGNAL,

  ADVANCE_VIEW_SIGNAL,
  BACK_VIEW_SIGNAL,

  DO_ON_SELECTION_SIGNAL,
  DO_ON_ITEM_AT_CURSOR_SIGNAL,

  URI_DROPPED_SIGNAL,

  LAST_SIGNAL
};


enum {
  POPUP_MENU_OPEN_SELECTION = 1,
  POPUP_MENU_OPEN_SELECTION_IN_TAB,
  POPUP_MENU_OPEN_SELECTION_WITH_BROWSER,
  POPUP_MENU_TOGGLE_HIDE_SELECTION,
  POPUP_MENU_KILL_SELECTION,
  POPUP_MENU_COPY_BOARD_URL,
  POPUP_MENU_OPEN_PROPERTIES
};


static GtkTreeViewClass *parent_class = NULL;
static int boardlist_view_signals[LAST_SIGNAL] =
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

#if GTK_MINOR_VERSION <= 2
static GtkItemFactory *popup_menu_item_factory = NULL;
static GtkItemFactory *popup_menu_item_factory2 = NULL;
#else
static GtkUIManager *ui_manager = NULL;
#endif
static GQuark node_info_id;

static char *list_entry_color[4] = { NULL, NULL, NULL, NULL };


static void
category_node_info_free(CategoryNodeInfo *info)
{
  gtk_tree_row_reference_free(info->reference);
  info->reference = NULL;
  G_FREE(info);
}


static CategoryNodeInfo *
ensure_node_info(GObject *object)
{
  CategoryNodeInfo *info = g_object_get_qdata(object, node_info_id);

  if (info == NULL)
    {
      info = G_NEW0(CategoryNodeInfo, 1);
      g_object_set_qdata_full(object, node_info_id,
			      info, (GDestroyNotify)category_node_info_free);
    }

  return info;
}


static void
boardlist_view_class_init(BoardlistViewClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;
  GtkBindingSet *binding_set;

  /* ݥåץåץ˥塼ν */
#if GTK_MINOR_VERSION <= 2
  GtkItemFactoryEntry menu_items[] =
    {
      {
	_("/_Open"),				/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION,		/* act */
	NULL					/* type */
      },
      {
	_("/Open Boards in _Tab"),		/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION_IN_TAB,	/* act */
	NULL					/* type */
      },
      {
	_("/Open Boards with Web Browser"),	/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION_WITH_BROWSER,	/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Toggle Hide(_D)"),			/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_TOGGLE_HIDE_SELECTION,	/* act */
	NULL					/* type */
      },
      {
	_("/Kill(_K)"),				/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_KILL_SELECTION,		/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Copy Board's URL"),			/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_COPY_BOARD_URL,		/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Open Properties"),			/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback,			/* callback_func */
	POPUP_MENU_OPEN_PROPERTIES,		/* act */
	NULL					/* type */
      },
    };
  int nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);

  GtkItemFactoryEntry menu_items2[] =
    {
      {
	_("/_Open"),				/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION,		/* act */
	NULL					/* type */
      },
      {
	_("/Open Boards in _Tab"),		/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION_IN_TAB,	/* act */
	NULL					/* type */
      },
      {
	_("/Open Boards with Web Browser"),	/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_OPEN_SELECTION_WITH_BROWSER,	/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Hide(_D)"),				/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_TOGGLE_HIDE_SELECTION,	/* act */
	NULL					/* type */
      },
      {
	_("/Kill(_K)"),				/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_KILL_SELECTION,		/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Copy Board's URL"),			/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_COPY_BOARD_URL,		/* act */
	NULL					/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	(char *)"<Separator>"
      },
      {
	_("/Open Properties"),			/* menu path */
	NULL,					/* accelerator */
	popup_menu_callback2,			/* callback_func */
	POPUP_MENU_OPEN_PROPERTIES,		/* act */
	NULL					/* type */
      },
    };
  int nmenu_items2 = sizeof(menu_items2) / sizeof(menu_items2[0]);

  popup_menu_item_factory = gtk_item_factory_new(GTK_TYPE_MENU,
						 "<BoardlistViewPopup>",
						 NULL);
  gtk_item_factory_create_items(popup_menu_item_factory,
				nmenu_items, menu_items, NULL);

  popup_menu_item_factory2 = gtk_item_factory_new(GTK_TYPE_MENU,
						 "<BoardlistViewPopup2>",
						 NULL);
  gtk_item_factory_create_items(popup_menu_item_factory2,
				nmenu_items2, menu_items2, NULL);
#else
  static GtkActionEntry popup_menu_entries[] =
    {
      {
	"OpenBoardlistItems", GTK_STOCK_OPEN,
	N_("_Open"), NULL,
	N_("Open Board or Category"),
	(GCallback)popup_menu_open_cb
      },
      {
	"OpenBoardsInTab", GTK_STOCK_OPEN,
	N_("_Open in _Tab"), NULL,
	N_("Open Board in Tab"),
	(GCallback)popup_menu_open_board_in_tab_cb
      },
      {
	"OpenBoardsWithWebBrowser", GTK_STOCK_EXECUTE,
	N_("Open with Web Browser"), NULL,
	N_("Open Board with Web Browser"),
	(GCallback)popup_menu_open_board_with_web_browser_cb
      },
      {
	"ToggleBoardlistItemHiddenStates", NULL,
	N_("Toggle Hide(_D)"), NULL,
	N_("Toggle Hidden States of Items in Selection"),
	(GCallback)popup_menu_toggle_hidden_states_cb
      },
      {
	"HideBoardlistItems", NULL,
	N_("Hide(_D)"), NULL,
	N_("Hide Items in Selection"),
	(GCallback)popup_menu_toggle_hidden_states_cb
      },
      {
	"KillItems", GTK_STOCK_DELETE,
	N_("_Kill"), NULL,
	N_("Kill Items in Selection"),
	(GCallback)popup_menu_kill_items_cb
      },
      {
	"CopyBoardURL", GTK_STOCK_COPY,
	N_("Copy Boards' URL"), NULL,
	N_("Copy Boards' URL"),
	(GCallback)popup_menu_copy_board_url_cb
      },
      {
	"OpenProperties", GTK_STOCK_PROPERTIES,
	N_("Open Properties"), NULL,
	N_("Open Properties"),
	(GCallback)popup_menu_open_properties_cb
      }
    };

  static const char *popup_menu_ui =
    "<ui>"
    "  <popup name=\"BoardlistPopup\">"
    "    <menuitem action=\"OpenBoardlistItems\"/>"
    "    <menuitem action=\"OpenBoardsInTab\"/>"
    "    <menuitem action=\"OpenBoardsWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"ToggleBoardlistItemHiddenStates\"/>"
    "    <menuitem action=\"KillItems\"/>"
    "    <separator/>"
    "    <menuitem action=\"CopyBoardURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"OpenProperties\"/>"
    "  </popup>"
    "  <popup name=\"SmartBoardlistPopup\">"
    "    <menuitem action=\"OpenBoardlistItems\"/>"
    "    <menuitem action=\"OpenBoardsInTab\"/>"
    "    <menuitem action=\"OpenBoardsWithWebBrowser\"/>"
    "    <separator/>"
    "    <menuitem action=\"HideBoardlistItems\"/>"
    "    <menuitem action=\"KillItems\"/>"
    "    <separator/>"
    "    <menuitem action=\"CopyBoardURL\"/>"
    "    <separator/>"
    "    <menuitem action=\"OpenProperties\"/>"
    "  </popup>"
    "</ui>";

  GError *error = NULL;
  GtkActionGroup *action_group
    = gtk_action_group_new("BoardlistViewPopupActions");
  ui_manager = gtk_ui_manager_new();

  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       popup_menu_entries,
			       G_N_ELEMENTS(popup_menu_entries),
			       ui_manager);
  gtk_ui_manager_insert_action_group(ui_manager, action_group, 0);

  if (!gtk_ui_manager_add_ui_from_string(ui_manager, popup_menu_ui, -1,
					 &error))
    {
      g_message("building menus failed: %s\n", error->message);
      g_error_free(error);
      exit(EXIT_FAILURE);
    }
#endif

  parent_class = g_type_class_peek_parent(klass);
  binding_set = gtk_binding_set_by_class(klass);

  /* GObject signals */
  o_class->finalize = boardlist_view_finalize;

  /* GtkObject signals */
  object_class->destroy = boardlist_view_destroy;

  gtk_widget_class_install_style_property(widget_class,
					  g_param_spec_string("fg_normal",
							      _("FG normal"),
							      _("Foreground color for ordinal list entry"),
							      NULL,
							      G_PARAM_READABLE));
  gtk_widget_class_install_style_property(widget_class,
					  g_param_spec_string("fg_hidden",
							      _("FG hidden"),
							      _("Foreground color for hidden list entry"),
							      "gray75",
							      G_PARAM_READABLE));
  gtk_widget_class_install_style_property(widget_class,
					  g_param_spec_string("bg_normal",
							      _("BG normal"),
							      _("Background color for ordinal list entry"),
							      NULL,
							      G_PARAM_READABLE));
  gtk_widget_class_install_style_property(widget_class,
					  g_param_spec_string("bg_hidden",
							      _("BG hidden"),
							      _("Background color for hidden list entry"),
							      NULL,
							      G_PARAM_READABLE));

  boardlist_view_signals[OPEN_THREADLIST_VIEW_SIGNAL] =
    g_signal_new("open_threadlist_view",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, open_threadlist_view),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT_BOOLEAN_BOOLEAN,
		 G_TYPE_NONE, 3,
		 OCHUSHA_TYPE_BULLETIN_BOARD,
		 G_TYPE_BOOLEAN,
		 G_TYPE_BOOLEAN);

  boardlist_view_signals[SELECT_CATEGORY_SIGNAL] =
    g_signal_new("select_category",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, select_category),
		 NULL, NULL,
		 ochusha_marshal_VOID__POINTER,
		 G_TYPE_NONE, 1,
		 G_TYPE_POINTER);
  boardlist_view_signals[CATEGORY_EXPANDED_SIGNAL] =
    g_signal_new("category_expanded",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, category_expanded),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 OCHUSHA_TYPE_BOARD_CATEGORY);
  boardlist_view_signals[CATEGORY_COLLAPSED_SIGNAL] =
    g_signal_new("category_collapsed",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, category_collapsed),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 OCHUSHA_TYPE_BOARD_CATEGORY);

  boardlist_view_signals[TOGGLE_HIDE_CATEGORY_SIGNAL] =
    g_signal_new("toggle_hide_category",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, toggle_hide_category),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 OCHUSHA_TYPE_BOARD_CATEGORY);
  boardlist_view_signals[TOGGLE_HIDE_BOARD_SIGNAL] =
    g_signal_new("toggle_hide_board",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, toggle_hide_board),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 OCHUSHA_TYPE_BULLETIN_BOARD);

  boardlist_view_signals[KILL_CATEGORY_SIGNAL] =
    g_signal_new("kill_category",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, kill_category),
		 NULL, NULL,
		 ochusha_marshal_BOOLEAN__OBJECT,
		 G_TYPE_BOOLEAN, 1,
		 OCHUSHA_TYPE_BOARD_CATEGORY);
  boardlist_view_signals[KILL_BOARD_SIGNAL] =
    g_signal_new("kill_board",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, kill_board),
		 NULL, NULL,
		 ochusha_marshal_BOOLEAN__OBJECT_OBJECT,
		 G_TYPE_BOOLEAN, 2,
		 OCHUSHA_TYPE_BOARD_CATEGORY,
		 OCHUSHA_TYPE_BULLETIN_BOARD);
  boardlist_view_signals[COPY_BOARD_URL_SIGNAL] =
    g_signal_new("copy_board_url",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, copy_board_url),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 OCHUSHA_TYPE_BULLETIN_BOARD);
  boardlist_view_signals[OPEN_PROPERTIES_SIGNAL] =
    g_signal_new("open_properties",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, open_properties),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 G_TYPE_OBJECT);

  boardlist_view_signals[ADVANCE_VIEW_SIGNAL] =
    g_signal_new("advance_view",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, advance_view),
		 NULL, NULL,
		 ochusha_marshal_BOOLEAN__OBJECT,
		 G_TYPE_BOOLEAN, 1,
		 OCHUSHA_TYPE_BULLETIN_BOARD);
  boardlist_view_signals[BACK_VIEW_SIGNAL] =
    g_signal_new("back_view",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(BoardlistViewClass, back_view),
		 NULL, NULL,
		 ochusha_marshal_BOOLEAN__OBJECT,
		 G_TYPE_BOOLEAN, 1,
		 OCHUSHA_TYPE_BULLETIN_BOARD);

  boardlist_view_signals[DO_ON_SELECTION_SIGNAL] =
    g_signal_new("do_on_selection",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(BoardlistViewClass, do_on_selection),
		 NULL, NULL,
		 ochusha_marshal_VOID__INT,
		 G_TYPE_NONE, 1,
		 G_TYPE_INT);
  boardlist_view_signals[DO_ON_ITEM_AT_CURSOR_SIGNAL] =
    g_signal_new("do_on_item_at_cursor",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(BoardlistViewClass, do_on_item_at_cursor),
		 NULL, NULL,
		 ochusha_marshal_VOID__ENUM,
		 G_TYPE_NONE, 1,
		 BOARDLIST_VIEW_ACTION_TYPE);

  boardlist_view_signals[URI_DROPPED_SIGNAL] =
    g_signal_new("uri_dropped",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(BoardlistViewClass, uri_dropped),
		 NULL, NULL,
		 ochusha_marshal_VOID__OBJECT_OBJECT_STRING_ENUM,
		 G_TYPE_NONE, 4,
		 OCHUSHA_TYPE_BOARD_CATEGORY,
		 OCHUSHA_TYPE_BULLETIN_BOARD,
		 G_TYPE_STRING,
		 GTK_TYPE_TREE_VIEW_DROP_POSITION);


  /* Key bindings */
  gtk_binding_entry_add_signal(binding_set, GDK_o, 0, "do_on_selection", 1,
			       G_TYPE_INT, POPUP_MENU_OPEN_SELECTION);
  gtk_binding_entry_add_signal(binding_set, GDK_t, 0, "do_on_selection", 1,
			       G_TYPE_INT, POPUP_MENU_OPEN_SELECTION_IN_TAB);
  gtk_binding_entry_add_signal(binding_set, GDK_d, 0, "do_on_item_at_cursor",
			       1, G_TYPE_INT, TOGGLE_HIDDEN_STATE);
  gtk_binding_entry_add_signal(binding_set, GDK_k, 0, "do_on_item_at_cursor",
			       1, G_TYPE_INT, KILL_ITEM);

  gtk_binding_entry_add_signal(binding_set, GDK_space, 0,
			       "do_on_item_at_cursor", 1,
			       G_TYPE_INT, ADVANCE_VIEW);
  gtk_binding_entry_add_signal(binding_set, GDK_BackSpace, 0,
			       "do_on_item_at_cursor", 1,
			       G_TYPE_INT, BACK_VIEW);

  gtk_binding_entry_add_signal(binding_set, GDK_Up, 0,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_DISPLAY_LINES,
			       G_TYPE_INT, -1);
  gtk_binding_entry_add_signal(binding_set, GDK_Down, 0,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_DISPLAY_LINES,
			       G_TYPE_INT, 1);
  gtk_binding_entry_add_signal(binding_set, GDK_p, GDK_CONTROL_MASK,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_DISPLAY_LINES,
			       G_TYPE_INT, -1);
  gtk_binding_entry_add_signal(binding_set, GDK_n, GDK_CONTROL_MASK,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_DISPLAY_LINES,
			       G_TYPE_INT, 1);
  gtk_binding_entry_add_signal(binding_set, GDK_v, GDK_CONTROL_MASK,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_PAGES,
			       G_TYPE_INT, 1);
  gtk_binding_entry_add_signal(binding_set, GDK_v, GDK_MOD1_MASK,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_PAGES,
			       G_TYPE_INT, -1);
  gtk_binding_entry_add_signal(binding_set, GDK_v, GDK_MOD4_MASK,
			       "move_cursor", 2,
			       G_TYPE_ENUM, GTK_MOVEMENT_PAGES,
			       G_TYPE_INT, -1);

  widget_class->enter_notify_event = boardlist_view_enter_notify;
  widget_class->button_press_event = boardlist_view_button_press;
  widget_class->button_release_event = boardlist_view_button_release;

  klass->open_threadlist_view = NULL;

  klass->select_category = NULL;
  klass->category_expanded = NULL;
  klass->category_collapsed = NULL;

  klass->toggle_hide_category = NULL;
  klass->toggle_hide_board = NULL;
  klass->kill_category = NULL;
  klass->kill_board = NULL;
  klass->copy_board_url = NULL;

  klass->advance_view = NULL;
  klass->back_view = NULL;

  klass->do_on_selection = boardlist_view_do_on_selection;
  klass->do_on_item_at_cursor = boardlist_view_do_on_item_at_cursor;

  node_info_id = g_quark_from_static_string("BoardlistView::CategoryNodeInfo");
}


enum
{
  BOARDLIST_CATEGORY_ROW,
  BOARDLIST_BOARD_ROW
};


enum
{
  BOARDLIST_LABEL = 0,
  BOARDLIST_ROW_TYPE,
  BOARDLIST_ROW_STATE,
  BOARDLIST_DATA,
  BOARDLIST_N_COLUMNS
};


enum
{
  TARGET_URI_LIST,
  TARGET_UTF8_STRING,
  TARGET_STRING,
  TARGET_TEXT,
  TARGET_COMPOUND_TEXT
};


static void
drag_data_received_cb(GtkWidget *widget, GdkDragContext *context,
		      int x, int y, GtkSelectionData *selection_data,
		      guint info, guint time, BoardlistView *view)
{
  int row_type = 0;
  gpointer data = NULL;
  GtkTreePath *path;
  GtkTreeViewDropPosition pos;
  OchushaBoardCategory *category = NULL;
  OchushaBulletinBoard *board = NULL;
  GtkTreeIter iter;
  const char *tmp_pos;
  char uri_buffer[PATH_MAX];

#if 0
  fprintf(stderr, "drag_data_received_cb: is_source=%d actions=0x%x suggested_action=0x%x action=0x%x time=%d\n",
	  context->is_source, context->actions, context->suggested_action,
	  context->action, time);
#endif
  if (!gtk_tree_view_get_dest_row_at_pos(&view->view, x, y, &path, &pos))
    {
      fprintf(stderr, "Couldn't know drop position.\n");
      return;
    }

  if (!gtk_tree_model_get_iter(view->model, &iter, path))
    {
      fprintf(stderr, "Couldn't get GtkTreeIter for path.\n");
      goto done;
    }

  gtk_tree_model_get(view->model, &iter,
		     BOARDLIST_ROW_TYPE, &row_type,
		     BOARDLIST_DATA, &data,
		     -1);

  switch (row_type)
    {
    case BOARDLIST_CATEGORY_ROW:
      category = OCHUSHA_BOARD_CATEGORY(data);
      break;

    case BOARDLIST_BOARD_ROW:
      board = OCHUSHA_BULLETIN_BOARD(data);
      if (gtk_tree_path_up(path)
	  && gtk_tree_model_get_iter(view->model, &iter, path))
	{
	  gtk_tree_model_get(view->model, &iter,
			     BOARDLIST_DATA, &category, -1);
	}
      break;
    }

  tmp_pos = (const char *)selection_data->data;
  while (tmp_pos != NULL)
    {
      size_t len;
      const char *tail_pos = strstr(tmp_pos, "\r\n");
      if (tail_pos == NULL)
	break;

      len = tail_pos - tmp_pos;
      if (len > 0 && len < PATH_MAX)
	{
	  memcpy(uri_buffer, tmp_pos, len);
	  uri_buffer[len] = '\0';
	  g_signal_emit(G_OBJECT(view),
			boardlist_view_signals[URI_DROPPED_SIGNAL],
			0,
			category, board, uri_buffer, pos);
	}
      tmp_pos = tail_pos + 2;
    }

  gtk_drag_finish(context, TRUE, FALSE, time);

 done:
  if (path != NULL)
    gtk_tree_path_free(path);
}


typedef struct _CollectArgs
{
  GtkTreeModel *model;
  GSList **list_p;
} CollectArgs;


static void
collect_drag_sources(gpointer list_data, gpointer user_data)
{
  GtkTreePath *path = (GtkTreePath *)list_data;
  CollectArgs *args = (CollectArgs *)user_data;
  GtkTreeModel *model = args->model;
  GSList **list_p = args->list_p;
  GtkTreeIter iter;
  gpointer data = NULL;

  if (!gtk_tree_model_get_iter(model, &iter, path))
    goto done;

  gtk_tree_model_get(model, &iter, BOARDLIST_DATA, &data, -1);
  if (data == NULL)
    goto done;

  *list_p = g_slist_append(*list_p, data);

 done:
  gtk_tree_path_free(path);
}


static void
append_uri(gpointer data, gpointer list_p)
{
  char **uri_list_p = (char **)list_p;

  if (OCHUSHA_IS_BULLETIN_BOARD(data))
    {
      if (*uri_list_p == NULL)
	*uri_list_p = g_strdup_printf("%s\r\n", ochusha_bulletin_board_get_base_url(OCHUSHA_BULLETIN_BOARD(data)));
      else
	{
	  char *prev_list = *uri_list_p;
	  *uri_list_p = g_strdup_printf("%s%s\r\n", prev_list, ochusha_bulletin_board_get_base_url(OCHUSHA_BULLETIN_BOARD(data)));
	  g_free(prev_list);
	}
    }
}


static void
append_utf8_text(gpointer data, gpointer text_p)
{
  gchar **utf8_text_p = (gchar **)text_p;

  if (OCHUSHA_IS_BULLETIN_BOARD(data))
    {
      OchushaBulletinBoard *board = OCHUSHA_BULLETIN_BOARD(data);
      if (*utf8_text_p == NULL)
	*utf8_text_p = g_strdup_printf(_("%s board\n%s\n"),
				       ochusha_bulletin_board_get_name(board),
				       ochusha_bulletin_board_get_base_url(board));
      else
	{
	  char *prev_text = *utf8_text_p;
	  *utf8_text_p = g_strdup_printf(_("%s%s board\n%s\n"),
					 prev_text,
					 ochusha_bulletin_board_get_name(board),
					 ochusha_bulletin_board_get_base_url(board));
	  g_free(prev_text);
	}
    }
  else if (OCHUSHA_BOARD_CATEGORY(data))
    {
      OchushaBoardCategory *cat = OCHUSHA_BOARD_CATEGORY(data);
      if (*utf8_text_p == NULL)
	*utf8_text_p = g_strdup_printf(_("Category %s\n"),
				       ochusha_board_category_get_name(cat));
      else
	{
	  char *prev_text = *utf8_text_p;
	  *utf8_text_p = g_strdup_printf(_("%sCategory %s\n"),
					 prev_text,
					 ochusha_board_category_get_name(cat));
	  g_free(prev_text);
	}
    }
}


static void
drag_data_get_cb(GtkWidget *widget, GdkDragContext *context,
		 GtkSelectionData *selection_data, guint info, guint time,
		 BoardlistView *view)
{
  GSList *data_list = NULL;
  GList *rows = gtk_tree_selection_get_selected_rows(view->selection, NULL);
  CollectArgs args = { view->model, &data_list };

  g_list_foreach(rows, collect_drag_sources, &args);
  g_list_free(rows);

  if (data_list != NULL)
    {
      switch (info)
	{
	case TARGET_URI_LIST:
	  {
	    char *uri_list = NULL;
	    g_slist_foreach(data_list, append_uri, &uri_list);
	    if (uri_list != NULL)
	      {
		gtk_selection_data_set(selection_data, selection_data->target,
				       8, uri_list, strlen(uri_list));
		g_free(uri_list);
	      }
	    break;
	  }

	case TARGET_UTF8_STRING:
	case TARGET_STRING:
	case TARGET_TEXT:
	case TARGET_COMPOUND_TEXT:
	  {
	    char *utf8_text = NULL;
	    g_slist_foreach(data_list, append_utf8_text, &utf8_text);
	    if (utf8_text != NULL)
	      {
		gtk_selection_data_set_text(selection_data, utf8_text, -1);
		g_free(utf8_text);
	      }
	    break;
	  }
	}
    }

  if (data_list != NULL)
    g_slist_free(data_list);

  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }
}


static void
boardlist_view_init(BoardlistView *view)
{
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;
  static const GtkTargetEntry drop_types[] =
    {
      { "text/uri-list", 0, TARGET_URI_LIST },
    };
  static int n_drop_types = sizeof(drop_types)/sizeof(drop_types[0]);

  static const GtkTargetEntry drag_types[] =
    {
      { "text/uri-list", 0, TARGET_URI_LIST },
      { "UTF8_STRING", 0, TARGET_UTF8_STRING },
      { "STRING", 0, 0 },
      { "TEXT", 0, 0 },
      { "COMPOUND_TEXT", 0, 0 }
    };
  static int n_drag_types = sizeof(drag_types)/sizeof(drag_types[0]);

  GTK_WIDGET_SET_FLAGS(view, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);

  view->selection = gtk_tree_view_get_selection(&view->view);
  gtk_tree_selection_set_mode(view->selection, GTK_SELECTION_MULTIPLE);

  g_signal_connect(G_OBJECT(view), "row-expanded",
		   G_CALLBACK(tree_view_row_expanded_cb), view);
  g_signal_connect(G_OBJECT(view), "row-collapsed",
		   G_CALLBACK(tree_view_row_collapsed_cb), view);
  view->last_event_path = NULL;

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes(_("Board List"), renderer,
						    "text", BOARDLIST_LABEL,
						    NULL);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  decorate_boardlist_entry,
					  view, NULL);
  gtk_tree_view_append_column(&view->view, column);

  /* D&Dμ¸ */
  gtk_tree_view_enable_model_drag_dest(&view->view,
				       drop_types, n_drop_types,
				       GDK_ACTION_COPY);
  g_signal_connect(view, "drag_data_received",
		   G_CALLBACK(drag_data_received_cb), view);

  gtk_tree_view_enable_model_drag_source(&view->view,
					 GDK_BUTTON1_MASK,
					 drag_types, n_drag_types,
					 GDK_ACTION_COPY);
  g_signal_connect(view, "drag_data_get",
		   G_CALLBACK(drag_data_get_cb), view);
}


static void
boardlist_view_finalize(GObject *object)
{
  BoardlistView *view = (BoardlistView *)object;

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "boardlist_view_finalize\n");
#endif
  if (view->model != NULL)
    {
      g_object_unref(view->model);	/* ȥ졼ʤΤ̵ */
      view->model = NULL;
    }

  if (G_OBJECT_CLASS(parent_class)->finalize)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
boardlist_view_destroy(GtkObject *object)
{
  BoardlistView *view = BOARDLIST_VIEW(object);
#if DEBUG_WIDGET_MOST
  fprintf(stderr, "boardlist_view_destroy\n");
#endif

  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


gboolean
boardlist_view_get_default_open_in_tab(BoardlistView *view)
{
  g_return_val_if_fail(IS_BOARDLIST_VIEW(view), FALSE);
  return view->default_open_in_tab;
}


void
boardlist_view_set_default_open_in_tab(BoardlistView *view,
				       gboolean default_open_in_tab)
{
  g_return_if_fail(IS_BOARDLIST_VIEW(view));
  view->default_open_in_tab = default_open_in_tab;
}


void
boardlist_view_expand_category(BoardlistView *view,
			       OchushaBoardCategory *category)
{
  gboolean expanded;
  CategoryNodeInfo *info = ensure_node_info(G_OBJECT(category));
  GtkTreePath *path;

  if (info->reference == NULL)
    return;	/* 줿ƥξ */

  path = gtk_tree_row_reference_get_path(info->reference);

  expanded = gtk_tree_view_expand_row(&view->view, path, FALSE);

  gtk_tree_path_free(path);

  if (expanded)
    g_signal_emit(G_OBJECT(view),
		  boardlist_view_signals[CATEGORY_EXPANDED_SIGNAL],
		  0,
		  category);
}


void
boardlist_view_collapse_category(BoardlistView *view,
				 OchushaBoardCategory *category)
{
  gboolean collapsed;
  CategoryNodeInfo *info = ensure_node_info(G_OBJECT(category));
  GtkTreePath *path;

  if (info->reference == NULL)
    return;	/* 줿ƥξ */

  path = gtk_tree_row_reference_get_path(info->reference);

  collapsed = gtk_tree_view_collapse_row(&view->view, path);

  gtk_tree_path_free(path);

  if (collapsed)
    g_signal_emit(G_OBJECT(view),
		  boardlist_view_signals[CATEGORY_COLLAPSED_SIGNAL],
		  0,
		  category);
}


enum {
  BOARDLIST_ROW_STATE_HIDDEN = 1,
  BOARDLIST_ROW_STATE_KILLED = 1 << 1,
};


static guint
encode_row_state(gboolean hidden, gboolean killed)
{
  return (hidden ? BOARDLIST_ROW_STATE_HIDDEN : 0)
    | (killed ? BOARDLIST_ROW_STATE_KILLED : 0);
}


typedef struct _DoOnSelectionArgs
{
  BoardlistView *view;
  int what_to_do;

  GSList *hidden_boards;
  GSList *hidden_categories;
} DoOnSelectionArgs;


static void
do_on_selection_helper(gpointer list_data, gpointer user_data)
{
  GtkTreePath *path = (GtkTreePath *)list_data;
  DoOnSelectionArgs *args = (DoOnSelectionArgs *)user_data;
  GtkTreeModel *model = args->view->model;
  int what_to_do = args->what_to_do;
  gpointer data = NULL;
  GtkTreeIter iter;

  if (!gtk_tree_model_get_iter(model, &iter, path))
    goto done;

  gtk_tree_model_get(model, &iter, BOARDLIST_DATA, &data, -1);
  if (data == NULL)
    goto done;	/* ͭʤǰΤ */

  switch (what_to_do)
    {
    case POPUP_MENU_OPEN_SELECTION:
      /* 2ܰʹߤ϶Ū˥֤ǳ */
      args->what_to_do = POPUP_MENU_OPEN_SELECTION_IN_TAB;

      /* FALL THROUGH */
    case POPUP_MENU_OPEN_SELECTION_IN_TAB:
      if (OCHUSHA_IS_BOARD_CATEGORY(data))
	gtk_tree_view_expand_row(&args->view->view, path, TRUE);
      else
	g_signal_emit(G_OBJECT(args->view),
		      boardlist_view_signals[OPEN_THREADLIST_VIEW_SIGNAL],
		      0,
		      data,
		      what_to_do == POPUP_MENU_OPEN_SELECTION_IN_TAB,
		      FALSE);
      break;

    case POPUP_MENU_OPEN_SELECTION_WITH_BROWSER:
      if (OCHUSHA_IS_BULLETIN_BOARD(data))
	g_signal_emit(G_OBJECT(args->view),
		      boardlist_view_signals[OPEN_THREADLIST_VIEW_SIGNAL],
		      0,
		      data,
		      FALSE,
		      TRUE);
      break;

    case POPUP_MENU_TOGGLE_HIDE_SELECTION:
      if (OCHUSHA_IS_BOARD_CATEGORY(data))
	{
	  OchushaBoardCategory *category = OCHUSHA_BOARD_CATEGORY(data);
	  g_signal_emit(G_OBJECT(args->view),
			boardlist_view_signals[TOGGLE_HIDE_CATEGORY_SIGNAL],
			0,
			category);
	  gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
			     BOARDLIST_ROW_STATE,
			     encode_row_state(category->hidden,
					      category->killed),
			     -1);
	  if (args->view->hide_hidden_items)
	    args->hidden_categories
	      = g_slist_append(args->hidden_categories,
			       gtk_tree_row_reference_new(model, path));
	}
      else if (OCHUSHA_IS_BULLETIN_BOARD(data))
	{
	  OchushaBulletinBoard *board = OCHUSHA_BULLETIN_BOARD(data);
	  g_signal_emit(G_OBJECT(args->view),
			boardlist_view_signals[TOGGLE_HIDE_BOARD_SIGNAL],
			0,
			board);
	  gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
			     BOARDLIST_ROW_STATE,
			     encode_row_state(board->hidden, board->killed),
			     -1);
	  if (args->view->hide_hidden_items)
	    args->hidden_boards
	      = g_slist_append(args->hidden_boards,
			       gtk_tree_row_reference_new(model, path));
	}
      break;

    case POPUP_MENU_KILL_SELECTION:
      if (OCHUSHA_IS_BOARD_CATEGORY(data))
	{
	  gboolean result = FALSE;
	  OchushaBoardCategory *category = OCHUSHA_BOARD_CATEGORY(data);
	  g_signal_emit(G_OBJECT(args->view),
			boardlist_view_signals[KILL_CATEGORY_SIGNAL],
			0,
			category,
			&result);
	  gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
			     BOARDLIST_ROW_STATE,
			     encode_row_state(category->hidden,
					      category->killed),
			     -1);
	  if (result)
	    args->hidden_categories
	      = g_slist_append(args->hidden_categories,
			       gtk_tree_row_reference_new(model, path));
	}
      else if (OCHUSHA_IS_BULLETIN_BOARD(data))
	{
	  gboolean result = FALSE;
	  GtkTreePath *category_path = gtk_tree_path_copy(path);
	  OchushaBulletinBoard *board = OCHUSHA_BULLETIN_BOARD(data);
	  OchushaBoardCategory *category = NULL;

	  if (gtk_tree_path_up(category_path))
	    {
	      GtkTreeIter category_iter;
	      if (gtk_tree_model_get_iter(model, &category_iter,
					  category_path))
		{
		  gtk_tree_model_get(model, &category_iter,
				     BOARDLIST_DATA, &category, -1);
		}
	    }
	  g_signal_emit(G_OBJECT(args->view),
			boardlist_view_signals[KILL_BOARD_SIGNAL],
			0,
			category,
			board,
			&result);
	  gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
			     BOARDLIST_ROW_STATE,
			     encode_row_state(board->hidden, board->killed),
			     -1);
	  if (result)
	    args->hidden_boards
	      = g_slist_append(args->hidden_boards,
			       gtk_tree_row_reference_new(model, path));
	  gtk_tree_path_free(category_path);
	}
      break;

    case POPUP_MENU_COPY_BOARD_URL:
      if (OCHUSHA_IS_BULLETIN_BOARD(data))
	g_signal_emit(G_OBJECT(args->view),
		      boardlist_view_signals[COPY_BOARD_URL_SIGNAL],
		      0,
		      data);
      break;

    case POPUP_MENU_OPEN_PROPERTIES:
      g_signal_emit(G_OBJECT(args->view),
		    boardlist_view_signals[OPEN_PROPERTIES_SIGNAL],
		    0,
		    data);
      break;
    }

 done:
  gtk_tree_path_free(path);
}


static void
remove_row(gpointer list_data, gpointer user_data)
{
  GtkTreeRowReference *reference = (GtkTreeRowReference *)list_data;
  GtkTreeStore *store = (GtkTreeStore *)user_data;
  GtkTreePath *path = gtk_tree_row_reference_get_path(reference);
  GtkTreeIter iter;

  if (!gtk_tree_model_get_iter(GTK_TREE_MODEL(store), &iter, path))
    goto done;

  gtk_tree_store_remove(store, &iter);

 done:
  if (path != NULL)
    gtk_tree_path_free(path);
  gtk_tree_row_reference_free(reference);
}


static void
boardlist_view_do_on_selection(BoardlistView *view, int what_to_do)
{
  GList *rows;
  DoOnSelectionArgs args = { view, what_to_do, NULL, NULL };
#if DEBUG_WIDGET_MOST
  fprintf(stderr, "boardlist_view_do_on_selection: view=%p what_to_do=%d\n",
	  view, what_to_do);
#endif

  rows = gtk_tree_selection_get_selected_rows(view->selection, NULL);
  g_list_foreach(rows, do_on_selection_helper, &args);
  g_list_free(rows);

  if (args.hidden_boards != NULL)
    {
      g_slist_foreach(args.hidden_boards, remove_row, view->model);
      g_slist_free(args.hidden_boards);
    }

  if (args.hidden_categories != NULL)
    {
      g_slist_foreach(args.hidden_categories, remove_row, view->model);
      g_slist_free(args.hidden_categories);
    }
}


static void
boardlist_view_do_on_item_at_cursor(BoardlistView *view,
				    BoardlistViewAction what_to_do)
{
  GtkTreeIter iter;
  GtkTreePath *path = NULL;
  gpointer data = NULL;

  gtk_tree_view_get_cursor(&view->view, &path, NULL);
  if (path == NULL)
    return;

  if (!gtk_tree_model_get_iter(view->model, &iter, path))
    goto done;

  gtk_tree_model_get(view->model, &iter, BOARDLIST_DATA, &data, -1);

  if (data != NULL)
    {
      gboolean result = FALSE;
      gtk_tree_selection_unselect_all(view->selection);
      (*parent_class->select_cursor_row)(&view->view, FALSE);

      switch (what_to_do)
	{
	case TOGGLE_HIDDEN_STATE:
	  boardlist_view_do_on_selection(view,
					 POPUP_MENU_TOGGLE_HIDE_SELECTION);
	  break;

	case KILL_ITEM:
	  boardlist_view_do_on_selection(view,
					 POPUP_MENU_KILL_SELECTION);
	  break;

	case ADVANCE_VIEW:
	  if (OCHUSHA_IS_BOARD_CATEGORY(data))
	    {
	      if (gtk_tree_view_row_expanded(&view->view, path))
		boardlist_view_collapse_category(view,
						 OCHUSHA_BOARD_CATEGORY(data));
	      else
		boardlist_view_expand_category(view,
					       OCHUSHA_BOARD_CATEGORY(data));
	      break;
	    }
	  boardlist_view_do_on_selection(view, POPUP_MENU_OPEN_SELECTION);
	  break;

	case BACK_VIEW:
	  if (OCHUSHA_IS_BOARD_CATEGORY(data))
	    {
	      if (gtk_tree_view_row_expanded(&view->view, path))
		boardlist_view_collapse_category(view,
						 OCHUSHA_BOARD_CATEGORY(data));
	      else
		boardlist_view_expand_category(view,
					       OCHUSHA_BOARD_CATEGORY(data));
	      break;
	    }
	  g_signal_emit(G_OBJECT(view),
			boardlist_view_signals[BACK_VIEW_SIGNAL],
			0,
			data,
			&result);
	  break;
	}
    }

 done:
  gtk_tree_path_free(path);
}


#if GTK_MINOR_VERSION <= 2
static void
popup_menu_callback(gpointer data, guint action, GtkWidget *widget)
{
  BoardlistView *view = gtk_item_factory_popup_data(popup_menu_item_factory);
#if DEBUG_WIDGET_MOST
  fprintf(stderr,
	  "<boardlist_view.c> popup_menu_callback: view=%p, action=%d\n",
	  view, action);
#endif
  boardlist_view_do_on_selection(view, action);
}


static void
popup_menu_callback2(gpointer data, guint action, GtkWidget *widget)
{
  BoardlistView *view = gtk_item_factory_popup_data(popup_menu_item_factory2);
#if DEBUG_WIDGET_MOST
  fprintf(stderr,
	  "<boardlist_view.c> popup_menu_callback: view=%p, action=%d\n",
	  view, action);
#endif
  boardlist_view_do_on_selection(view, action);
}
#else
static void
popup_menu_open_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_OPEN_SELECTION);
}


static void
popup_menu_open_board_in_tab_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_OPEN_SELECTION_IN_TAB);
}


static void
popup_menu_open_board_with_web_browser_cb(GtkAction *action,
					  GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_OPEN_SELECTION_WITH_BROWSER);
}


static void
popup_menu_toggle_hidden_states_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_TOGGLE_HIDE_SELECTION);
}


static void
popup_menu_kill_items_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_KILL_SELECTION);
}


static void
popup_menu_copy_board_url_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_COPY_BOARD_URL);
}


static void
popup_menu_open_properties_cb(GtkAction *action, GtkUIManager *manager)
{
  BoardlistView *view = ochusha_ui_manager_get_popup_data(manager);
  boardlist_view_do_on_selection(view, POPUP_MENU_OPEN_PROPERTIES);
}
#endif


static void
popup_menu_destructed_cb(gpointer data)
{
#if DEBUG_WIDGET_MOST
  BoardlistView *view = BOARDLIST_VIEW(data);
  fprintf(stderr, "<boardlist_view.c> popup_menu_destructed_cb: data=%p\n",
	  view);
#endif
}


GtkWidget *
boardlist_view_new(void)
{
  return GTK_WIDGET(g_object_new(BOARDLIST_VIEW_TYPE, NULL));
}


static gboolean
boardlist_view_enter_notify(GtkWidget *widget, GdkEventCrossing *event)
{
  OchushaApplication *application = g_object_get_qdata(G_OBJECT(widget),
						       application_id);
  if (application != NULL && application->focus_follow_mouse)
    gtk_widget_grab_focus(widget);

  if (GTK_WIDGET_CLASS(parent_class)->enter_notify_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->enter_notify_event)(widget,
								 event);
  return FALSE;
}


static gboolean
boardlist_view_button_press(GtkWidget *widget, GdkEventButton *event)
{
  BoardlistView *view;
  GtkTreeIter iter;
  GtkTreePath *path = NULL;
  gboolean result = FALSE;

  view = BOARDLIST_VIEW(widget);

  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (!gtk_tree_view_get_path_at_pos(&view->view, event->x, event->y, &path,
				     NULL, NULL, NULL))
    goto done;

  if (!gtk_tree_model_get_iter(view->model, &iter, path))
    goto done;

  if (event->button == 3)
    {
#if GTK_MINOR_VERSION <= 2
      if (view->hide_hidden_items)
	gtk_item_factory_popup_with_data(popup_menu_item_factory2,
					 view, popup_menu_destructed_cb,
					 event->x_root, event->y_root,
					 event->button,
					 gtk_get_current_event_time());
      else
	gtk_item_factory_popup_with_data(popup_menu_item_factory,
					 view, popup_menu_destructed_cb,
					 event->x_root, event->y_root,
					 event->button,
					 gtk_get_current_event_time());
#else
      if (view->hide_hidden_items)
	ochusha_ui_manager_popup_with_data(ui_manager,
					   "/SmartBoardlistPopup",
					   view, popup_menu_destructed_cb,
					   event->x_root, event->y_root,
					   event->button,
					   gtk_get_current_event_time());
      else
	ochusha_ui_manager_popup_with_data(ui_manager,
					   "/BoardlistPopup",
					   view, popup_menu_destructed_cb,
					   event->x_root, event->y_root,
					   event->button,
					   gtk_get_current_event_time());
#endif

      if (!gtk_tree_selection_iter_is_selected(view->selection, &iter))
	{
	  /* XXX: ȡårow򤵤줿֤ˤʤ롣
	   *      CTRLSHIFT򲡤ʤξˤϰ㤦񤤤
	   *      ٤ȿˡĤʤΤŶ
	   *      DnDݡΥСǤϡå줿row
	   *      ƤʤäˤϤΥʥϥɥ餬FALSE
	   *      ֤ȤˤꡢGtkTreeView˽Ǥä
	   *      DnDݡȻƱˡȡܥ󥯥åǤ
	   *      DnDǧƤʤϤʤΤDnD⡼ɤäƤޤ
	   *      ݥåץåץ˥塼̥ɥʤΤǡ̯ʸ
	   *      롣
	   */
	  gtk_tree_view_set_cursor(&view->view, path, NULL, FALSE);
	}
      result = TRUE;
      goto done;
    }

  if ((event->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) != 0)
    goto done;

  view->last_event_path = path;
  path = NULL;

 done:

  if (path != NULL)
    gtk_tree_path_free(path);

  if (!result && GTK_WIDGET_CLASS(parent_class)->button_press_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->button_press_event)(widget,
								 event);
  return TRUE;
}


static gboolean
boardlist_view_button_release(GtkWidget *widget, GdkEventButton *event)
{
  BoardlistView *view = (BoardlistView *)widget;
  GtkTreePath *path = NULL;
  gboolean check_expander = FALSE;
  gboolean expanded = FALSE;
  gboolean result;

  g_return_val_if_fail(IS_BOARDLIST_VIEW(view), FALSE);
  g_return_val_if_fail(event != NULL, FALSE);

  if (event->button != 3 && view->last_event_path != NULL)
    {
      int row_type;
      gpointer data = NULL;
      GtkTreeIter iter;
      if (!gtk_tree_view_get_path_at_pos(&view->view, event->x, event->y,
					 &path,
					 NULL, NULL, NULL))
	goto done;

      if (gtk_tree_path_compare(path, view->last_event_path) != 0)
	goto done;

      if (!gtk_tree_model_get_iter(view->model, &iter, path))
	goto done;

      gtk_tree_model_get(view->model, &iter,
			 BOARDLIST_ROW_TYPE, &row_type,
			 BOARDLIST_DATA, &data,
			 -1);

      switch (row_type)
	{
	case BOARDLIST_CATEGORY_ROW:
	  expanded = gtk_tree_view_row_expanded(&view->view, path);
	  g_signal_emit(G_OBJECT(view),
			boardlist_view_signals[SELECT_CATEGORY_SIGNAL],
			0,
			data);
	  check_expander = TRUE;
	  break;

	case BOARDLIST_BOARD_ROW:
	  g_signal_emit(G_OBJECT(view),
			boardlist_view_signals[OPEN_THREADLIST_VIEW_SIGNAL],
			0,
			data,
			view->default_open_in_tab
			? event->button == 1 : event->button == 2,
			FALSE);
	  break;

	default:
	  fprintf(stderr, "GtkTreeModel broken.\n");
	}
    }

 done:
  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (GTK_WIDGET_CLASS(parent_class)->button_release_event != NULL)
    result = (*GTK_WIDGET_CLASS(parent_class)->button_release_event)(widget,
								     event);
  else
    result = FALSE;

  if (check_expander)
    {
      if (gtk_tree_view_row_expanded(&view->view, path) == expanded)
	{
	  if (expanded)
	    gtk_tree_view_collapse_row(&view->view, path);
	  else
	    gtk_tree_view_expand_row(&view->view, path, FALSE);
	}
    }

  if (path != NULL)
    gtk_tree_path_free(path);

  return result;
}


static void
tree_view_row_expanded_cb(GtkTreeView *tree_view,
			  GtkTreeIter *iter, GtkTreePath *path,
			  BoardlistView *view)
{
  OchushaBoardCategory *category;
#if DEBUG_WIDGET_MOST
  gchar *name;
#endif

  g_return_if_fail(tree_view != NULL);
  g_return_if_fail(view != NULL);

#if DEBUG_WIDGET_MOST
  gtk_tree_model_get(view->model, iter,
		     BOARDLIST_LABEL, &name,
		     BOARDLIST_DATA, &category,
		     -1);
  fprintf(stderr, "row_expanded_cb: %s\n", name);
  G_FREE(name);
#else
  gtk_tree_model_get(view->model, iter,
		     BOARDLIST_DATA, &category,
		     -1);
#endif

  if (category != NULL)
    {
      g_signal_emit(G_OBJECT(view),
		    boardlist_view_signals[CATEGORY_EXPANDED_SIGNAL],
		    0,
		    category);
    }
}


static void
tree_view_row_collapsed_cb(GtkTreeView *tree_view,
			   GtkTreeIter *iter, GtkTreePath *path,
			   BoardlistView *view)
{
  OchushaBoardCategory *category;
#if DEBUG_WIDGET_MOST
  gchar *name;
#endif

  g_return_if_fail(tree_view != NULL);
  g_return_if_fail(view != NULL);

#if DEBUG_WIDGET_MOST
  gtk_tree_model_get(view->model, iter,
		     BOARDLIST_LABEL, &name,
		     BOARDLIST_DATA, &category,
		     -1);
  fprintf(stderr, "row_collapsed_cb: %s\n", name);
  G_FREE(name);
#else
  gtk_tree_model_get(view->model, iter,
		     BOARDLIST_DATA, &category,
		     -1);
#endif

  if (category != NULL)
    {
      g_signal_emit(G_OBJECT(view),
		    boardlist_view_signals[CATEGORY_COLLAPSED_SIGNAL],
		    0,
		    category);
    }
}


static const char *boardlist_view_get_color(BoardlistView *view,
					    guint color_code);

static const char *
find_replacement_color(BoardlistView *view, const char *color)
{
  const char *result_color;
  GdkColor *gdk_color;
  char name_buf[8];
  g_return_val_if_fail(color != NULL && *color == '<', NULL);

  if (strncmp(color + 1, "fg_", 3) == 0)
    {
      if (strcmp(color + 4, "normal>") == 0)
	result_color = boardlist_view_get_color(view, 0);
      else if (strcmp(color + 4, "hidden>") == 0)
	result_color = boardlist_view_get_color(view, 1);
      else
	return NULL;
      if (result_color != NULL)
	return result_color;

      gdk_color = &GTK_WIDGET(view)->style->text[GTK_STATE_NORMAL];
      snprintf(name_buf, 8, "#%02x%02x%02x",
	       gdk_color->red / 256, gdk_color->green / 256,
	       gdk_color->blue / 256);
      return g_strdup(name_buf);	/* Ĥϥ꡼뤱̵ */
    }
  else if (strncmp(color + 1, "bg_", 3) == 0)
    {
      if (strcmp(color + 4, "normal>") == 0)
	result_color = boardlist_view_get_color(view, 2);
      else if (strcmp(color + 4, "hidden>") == 0)
	result_color = boardlist_view_get_color(view, 3);
      else
	return NULL;
      if (result_color != NULL)
	return result_color;

      gdk_color = &GTK_WIDGET(view)->style->base[GTK_STATE_NORMAL];
      snprintf(name_buf, 8, "#%02x%02x%02x",
	       gdk_color->red / 256, gdk_color->green / 256,
	       gdk_color->blue / 256);
      return g_strdup(name_buf);	/* Ĥϥ꡼뤱̵ */
    }

  return NULL;
}


static const char *
boardlist_view_get_color(BoardlistView *view, guint color_code)
{
  const char *color;
  char *tmp_color;

  g_return_val_if_fail(color_code < 4, NULL);

  if (list_entry_color[color_code] != NULL)
    {
      color = list_entry_color[color_code];
      if (color != NULL && *color == '\0')
	return NULL;
      return color;
    }

  switch (color_code)
    {
    case 0:
      gtk_widget_style_get(GTK_WIDGET(view), "fg_normal", &color, NULL);
      break;
    case 1:
      gtk_widget_style_get(GTK_WIDGET(view), "fg_hidden", &color, NULL);
      break;
    case 2:
      gtk_widget_style_get(GTK_WIDGET(view), "bg_normal", &color, NULL);
      break;
    case 3:
      gtk_widget_style_get(GTK_WIDGET(view), "bg_hidden", &color, NULL);
      break;
    default:
      color = NULL;
    }

  if (color == NULL)
    return color;

  if (*color == '\0')
    return NULL;

  if (*color != '<')
    return color;

  list_entry_color[color_code] = G_STRDUP("");

  color = find_replacement_color(view, color);

  if (color == NULL)
    return NULL;

  tmp_color = G_STRDUP(color);
  G_FREE(list_entry_color[color_code]);

  list_entry_color[color_code] = tmp_color;
  if (*tmp_color == '\0')
    return NULL;
  return tmp_color;
}


static void
decorate_boardlist_entry(GtkTreeViewColumn *tree_column,
			 GtkCellRenderer *renderer, GtkTreeModel *tree_model,
			 GtkTreeIter *iter, gpointer view)
{
  guint state;
  const char *color;

  g_return_if_fail(gtk_tree_store_iter_is_valid(GTK_TREE_STORE(tree_model),
						iter));
  gtk_tree_model_get(tree_model, iter,
		     BOARDLIST_ROW_STATE, &state, -1);

  if (!GTK_IS_CELL_RENDERER_TEXT(renderer))
    return;

  if ((state & BOARDLIST_ROW_STATE_HIDDEN) != 0)
    {
      if ((state & BOARDLIST_ROW_STATE_KILLED) == 0)
	color = boardlist_view_get_color(view, 1);
      else
	color = boardlist_view_get_color(view, 3);
      g_object_set((GObject *)renderer,
		   "strikethrough", TRUE,
		   "strikethrough_set", TRUE,
		   "cell_background", boardlist_view_get_color(view, 3),
		   "foreground", color,
		   NULL);
    }
  else
    {
      if ((state & BOARDLIST_ROW_STATE_KILLED) == 0)
	color = boardlist_view_get_color(view, 0);
      else
	color = boardlist_view_get_color(view, 2);
      g_object_set((GObject *)renderer,
		   "strikethrough", FALSE,
		   "strikethrough_set", FALSE,
		   "cell_background", boardlist_view_get_color(view, 2),
		   "foreground", color,
		   NULL);
    }
}


typedef struct _AppendBoardArgs
{
  GtkTreeIter *category_iter;
  GtkTreeStore *store;
  gboolean really_hide_hidden_items;
} AppendBoardArgs;


static void
append_board(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  AppendBoardArgs *args = (AppendBoardArgs *)user_data;
  GtkTreeIter iter;

  if (board->hidden && args->really_hide_hidden_items)
    return;

  gtk_tree_store_append(args->store, &iter, args->category_iter);
  gtk_tree_store_set(args->store, &iter,
		     BOARDLIST_LABEL, board->name,
		     BOARDLIST_ROW_TYPE, BOARDLIST_BOARD_ROW,
		     BOARDLIST_ROW_STATE, encode_row_state(board->hidden,
							   board->killed),
		     BOARDLIST_DATA, board,
		     -1);
}


static void
append_category(gpointer data, gpointer user_data)
{
  OchushaBoardCategory *category = (OchushaBoardCategory *)data;
  CategoryNodeInfo *info = ensure_node_info(G_OBJECT(category));
  AppendBoardArgs *args = (AppendBoardArgs *)user_data;
  GtkTreeIter iter;
  GtkTreePath *path;

  if (category->hidden && args->really_hide_hidden_items)
    return;

  gtk_tree_store_append(args->store, &iter, NULL);
  gtk_tree_store_set(args->store, &iter,
		     BOARDLIST_LABEL, category->name,
		     BOARDLIST_ROW_TYPE, BOARDLIST_CATEGORY_ROW,
		     BOARDLIST_ROW_STATE, encode_row_state(category->hidden,
							   category->killed),
		     BOARDLIST_DATA, category,
		     -1);
  path = gtk_tree_model_get_path(GTK_TREE_MODEL(args->store), &iter);
  info->reference = gtk_tree_row_reference_new(GTK_TREE_MODEL(args->store),
					       path);
  gtk_tree_path_free(path);

  args->category_iter = &iter;

  g_slist_foreach(category->board_list, append_board, args);
}


void
boardlist_view_open(BoardlistView *view, GSList *category_list,
		    gboolean really_hide_hidden_items)
{
  GtkTreeStore *store = gtk_tree_store_new(BOARDLIST_N_COLUMNS,
		/* tree view label */	   G_TYPE_STRING,
		/* row type */		   G_TYPE_INT,
		/* row state */		   G_TYPE_INT,
		/* data */		   G_TYPE_POINTER);
  		/* (OchushaBoardCategory *)  (OchushaBulletinBoard *) */
  AppendBoardArgs args = { NULL, store, really_hide_hidden_items };
  view->hide_hidden_items = really_hide_hidden_items;

  g_slist_foreach(category_list, append_category, &args);

  gtk_tree_view_set_model(&view->view, GTK_TREE_MODEL(store));

  if (view->model != NULL)
    g_object_unref(view->model);	/* ȥ졼ʤΤ̵ */
  view->model = GTK_TREE_MODEL(store);
}


void
boardlist_view_add_category(BoardlistView *view,
			    OchushaBoardCategory *category,
			    OchushaBoardCategory *prev_category)
{
  CategoryNodeInfo *info;
  GtkTreeIter iter;
  GtkTreeIter prev_iter;
  GtkTreeIter *prev_iter_p = NULL;
  GtkTreePath *path;

  g_return_if_fail(IS_BOARDLIST_VIEW(view));
  g_return_if_fail(OCHUSHA_IS_BOARD_CATEGORY(category));

  info = ensure_node_info(G_OBJECT(category));
  g_return_if_fail(info != NULL);
  if (OCHUSHA_IS_BOARD_CATEGORY(prev_category))
    {
      CategoryNodeInfo *prev_info = ensure_node_info(G_OBJECT(prev_category));
      path = gtk_tree_row_reference_get_path(prev_info->reference);
      if (path != NULL)
	{
	  if (gtk_tree_model_get_iter(view->model, &prev_iter, path))
	    prev_iter_p = &prev_iter;
	  gtk_tree_path_free(path);
	}
    }

  gtk_tree_store_insert_after(GTK_TREE_STORE(view->model), &iter,
			      NULL, prev_iter_p);
  gtk_tree_store_set(GTK_TREE_STORE(view->model), &iter,
		     BOARDLIST_LABEL, category->name,
		     BOARDLIST_ROW_TYPE, BOARDLIST_CATEGORY_ROW,
		     BOARDLIST_ROW_STATE, encode_row_state(category->hidden,
							   category->killed),
		     BOARDLIST_DATA, category,
		     -1);
  path = gtk_tree_model_get_path(view->model, &iter);
  if (info->reference != NULL)
    gtk_tree_row_reference_free(info->reference);
  info->reference = gtk_tree_row_reference_new(view->model, path);
  gtk_tree_path_free(path);
}


void
boardlist_view_remove_category(BoardlistView *view,
			       OchushaBoardCategory *category)
{
  CategoryNodeInfo *info;

  g_return_if_fail(IS_BOARDLIST_VIEW(view));
  g_return_if_fail(OCHUSHA_IS_BOARD_CATEGORY(category));

  info = ensure_node_info(G_OBJECT(category));
  g_return_if_fail(info->reference != NULL);
  remove_row(info->reference, view->model);
  info->reference = NULL;
}


void
boardlist_view_add_board(BoardlistView *view, OchushaBulletinBoard *board,
			 OchushaBoardCategory *category,
			 OchushaBulletinBoard *prev_board)
{
  CategoryNodeInfo *info;
  GtkTreeIter category_iter;
  GtkTreePath *path;

  g_return_if_fail(IS_BOARDLIST_VIEW(view));
  g_return_if_fail(OCHUSHA_IS_BOARD_CATEGORY(category));
  g_return_if_fail(OCHUSHA_IS_BULLETIN_BOARD(board));

  info = ensure_node_info(G_OBJECT(category));
  g_return_if_fail(info != NULL || info->reference != NULL);

  path = gtk_tree_row_reference_get_path(info->reference);
  g_return_if_fail(path != NULL);

  if (gtk_tree_model_get_iter(view->model, &category_iter, path))
    {
      GtkTreeIter prev_iter;
      GtkTreeIter *prev_iter_p = NULL;
      GtkTreeIter iter;

      if (OCHUSHA_IS_BULLETIN_BOARD(prev_board))
	{
	  /* prev_iter_pν */
	  if (gtk_tree_model_iter_children(view->model, &prev_iter,
					   &category_iter))
	    {
	      do
		{
		  OchushaBulletinBoard *data;
		  gtk_tree_model_get(view->model, &prev_iter,
				     BOARDLIST_DATA, &data, -1);
		  if (data == prev_board)
		    {
		      prev_iter_p = &prev_iter;
		      break;
		    }
		} while (gtk_tree_model_iter_next(view->model, &prev_iter));
	    }
	}

      gtk_tree_store_insert_after(GTK_TREE_STORE(view->model), &iter,
				  &category_iter, prev_iter_p);
      gtk_tree_store_set(GTK_TREE_STORE(view->model), &iter,
			 BOARDLIST_LABEL, board->name,
			 BOARDLIST_ROW_TYPE, BOARDLIST_BOARD_ROW,
			 BOARDLIST_ROW_STATE, encode_row_state(board->hidden,
							       board->killed),
			 BOARDLIST_DATA, board,
			 -1);
    }

  gtk_tree_path_free(path);
}


void
boardlist_view_remove_board(BoardlistView *view,
			    OchushaBoardCategory *category,
			    OchushaBulletinBoard *board)
{
  CategoryNodeInfo *info;
  GtkTreeIter category_iter;
  GtkTreePath *path;

  g_return_if_fail(IS_BOARDLIST_VIEW(view));
  g_return_if_fail(OCHUSHA_IS_BOARD_CATEGORY(category));
  g_return_if_fail(OCHUSHA_IS_BULLETIN_BOARD(board));

  info = ensure_node_info(G_OBJECT(category));
  g_return_if_fail(info != NULL || info->reference != NULL);

  path = gtk_tree_row_reference_get_path(info->reference);
  g_return_if_fail(path != NULL);

  if (gtk_tree_model_get_iter(view->model, &category_iter, path))
    {
      GtkTreeIter iter;
      if (gtk_tree_model_iter_children(view->model, &iter, &category_iter))
	{
	  do
	    {
	      OchushaBulletinBoard *data;
	      gtk_tree_model_get(view->model, &iter,
				 BOARDLIST_DATA, &data, -1);
	      if (data == board)
		{
		  gtk_tree_store_remove(GTK_TREE_STORE(view->model), &iter);
		  break;
		}
	    } while (gtk_tree_model_iter_next(view->model, &iter));
	}
    }

  gtk_tree_path_free(path);
}

